function p = compute_p1w_PR(t,G,q,W)
% function p = compute_p1w_PR(t,G,q,W)
%
% compute numerator of reduced system (positive real, ellipsoid algorithm)
%
% t --- discrete-time frequencies
% G --- frequency response of the original system
% q --- denominator of the reduced system in decreasing order
% weigthed version of compute_p1

% preprocessing
t = t(:);
G = G(:);
nt = size(t,1);
q = q(:)';
m = size(q,2)-1;    % order of reduced system
if nargin < 4
    W = ones(nt,1);
else
    W = W(:);
end

% build some necessary matrices
h = (4*m+2)*max(abs(G));
x0 = zeros(m+1,1);                       % initial guess: p(z) = 0
M = 1e6;
H0 = diag(repmat(M,1,m+1));              % initial ellipsoid
Gn = G/h;                                % normalized samples as a column vector

z = exp(j*t);
zi = exp(-j*t);
qz = polyval(q,z);
qzi = polyval(q,zi);
Gqz2 = Gn.*abs(qz).^2;
P = (repmat(z,1,m+1).^repmat((m:-1:0),nt,1)).*repmat(qzi,1,m+1);

r = 0.95;
max_iter = 60000;

% call ellip_compute_p1w_PR
[p,y] = ellip_compute_p1w_PR(q,(abs(qz).^2)./W,Gqz2,P,H0,x0,r,max_iter);
p = p'*h;

function [xb,yb] = ellip_compute_p1w_PR(q,qz2,Gqz2,P,H0,x0,r,max_iter)
% function [xb,yb] = ellip_compute_p1w_PR(q,qz2,Gqz2,P,H0,x0,r,max_iter)
%
% ellipsoid algorithm for solving positive real numerator of reduced system
% problem
% min_p |p/q-G| : p(z)q(1/z)+p(1/z)q(z) > 0 for all |z| = 1
%
% Inputs:
% q --- denominator of the reduced system
% qz2 --- abs(q(z))^2 at evaluation points
% Gqz2 --- G(z)*abs(q(z))^2 at evaluation points
% P --- p(z)*q(1/z) = P*x at evaluation points
% H0 --- initial ellipsoid = {x : |inv(H0)*(x-x0)| <= 1
% x0 --- center of the initial ellipsoid
% r --- superoptimal level
% max_iter --- maximum allowable iterations
%
% Outputs:
% x --- center of the final ellipsoid
% y --- cost associated with x

% parameters
if nargin < 7
    r = 0.9;
end
if nargin < 8
    max_iter = 30000;
end

% initialization
xb = x0;    % current best x
yb = max(abs(P*xb-Gqz2)./qz2);    % current best F(x)
xk = x0;    % current of the current ellipsoid Ek
Hk = H0;    % matrix of the current ellipsoid Ek
LB_hist = zeros(max_iter,1);
cut_hist = zeros(max_iter,1);

% main iteration
for iter = 1:max_iter
    y = r*yb;   % current superoptimal level
    % add superoptimal cut
    tic;
    [L,ds] = obj_cut_compute_p1(qz2,Gqz2,P,xk,y);
    if ds ~= -1 % add superoptimal cut and update ellipsoid (or verify that the ellipsoid is empty)
        % check whether ellipsoid is empty or not
        nrm = norm(L*Hk);
        if nrm <= ds
            return
        end
        [Hk,xk]=ellips_cut1(Hk,xk,L,ds/nrm);
        cut_hist(iter) = 1;
    else
        % check feasiblity of xk (positive-realness)
        [L,ds] = PR_check(xk,q);
        if ds == -inf   % xk is feasible and upadate xb (yb)
            xb = xk;
            yb = max(abs(P*xk-Gqz2)./qz2);
        else    % xk is infeasible, check feasibility of xt = 0.5*(xb+xk)
            xt = 0.5*(xb+xk);
            [L1,ds1] = PR_check(xt,q);
            if ds1 ~= -inf  % xt is infeasible, hence add feasibility cut
                L = L1;
                ds = L*(xt-xk)+ds1;
                nrm = norm(L*Hk);
                if nrm <= ds
                    return
                else
                    [Hk,xk] = ellips_cut1(Hk,xk,L,ds/nrm);
                    cut_hist(iter) = -1;
                end
            else    % xt is feasible, add feasibility cut of xk
                if max(abs(P*xt-Gqz2)./qz2) < yb
                    xb = xt;
                    yb = max(abs(P*xt-Gqz2)./qz2);
                else
                % check whether ellipsoid is empty or not
                nrm = norm(L*Hk);
                if nrm <= ds
                    return
                else
                    % add feasibility cut
                    nrm = norm(L*Hk);
                    [Hk,xk] = ellips_cut1(Hk,xk,L,ds/nrm);
                    cut_hist(iter) = -1;
                end
            end
            end
        end
    end
    LB_hist(iter) = yb;
end

% return output due to iteration limit
disp('exit due to iteration limit!')

function [L,ds] = obj_cut_compute_p1(qz2,Gqz2,P,x,y)
% function [L,ds] = obj_cut_compute_p1(qz2,Gqz2,P,x,y)
%
% objective cut function

m = length(x)-1;
nmd = abs(P*x-Gqz2)-y*qz2;
[nmd_max,max_ind] = max(nmd);
if nmd_max < 0
    L = zeros(1,m+1);
    ds = -1;
else
    pm = real([P(max_ind,:) -Gqz2(max_ind)]*conj(sign(P(max_ind,:)*x-Gqz2(max_ind))));
    L = -pm(1:end-1);
    ds = pm(end)-y*qz2(max_ind)-L*x;
    if ds < 0
        ds = 1e-14;
    end
end

function [L,ds] = PR_check(x,q)
% funciton [L,ds] = PR_check(x,q)
%
% positive-realness check

tol = 1e-6;
m = length(x)-1;
b = conv(x,q(end:-1:1))+conv(x(end:-1:1),q);
if b(m+1) < 0   % b_0 should be positive
    L = 2*q;
    ds = -dot(L,x);
    return;
else    % check if b(t) > 0 for all t
    Zb = roots(b);
    [d,ind] = min(abs(abs(Zb)-1));
    if d <= tol % b(t) = 0 for some t
        t = angle(Zb(ind));
        H = hankel(q);
        T = toeplitz([q(1) ; zeros(m,1)],q);
        L = [1 2*cos((1:m)*t)]*(H+T);
        ds = -dot(L,x);
        return;
    else    % b(t) > 0 for all t (i.e. positive real)
        L = zeros(1,m+1);
        ds = -inf;
    end
end